#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <stdio.h>
#include <malloc.h>
#include "glATI.h"

#define GTC_ID (('C'<<16)+('T'<<8)+'G')

HWND hWnd=NULL;
HDC hDC=NULL;
HGLRC hRC=NULL;

char *szAppName="OpenGL";

RECT WindowRect;

int Width=160, Height=120;
BOOL Done=FALSE;
unsigned int compressed_texture;
int width, height, format;
char *file;

typedef struct
{
	short X_Origin;
	short Y_Origin;
	short Width;
	short Height;
	unsigned char Depth;
	unsigned char Image_Descriptor;
} TGA_ImageSpec_t;

typedef struct
{
	unsigned char ID;
	unsigned char ColorMapType;
	unsigned char ImageType;
	TGA_ImageSpec_t ImageSpec;
	unsigned char *Data;
} TGA_t;

typedef struct
{
	int ID;
	unsigned int Width;
	unsigned int Height;
	unsigned int Format;
	unsigned int Size;
	unsigned char *Data;
} GTC_t;

PFNGLGETCOMPRESSEDTEXIMAGEARBPROC glGetCompressedTexImageARB=NULL;

LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void Render(void);
BOOL Init(void);
BOOL Create(void);
void Destroy(void);
void Resize(int Width, int Height);
BOOL IsExtensionSupported(const char *extension);
BOOL TGA_GetData(char *file, TGA_t *TGA);
unsigned int TGA_UploadImage(char *file);
void GTC_WriteImage(char *file, unsigned int TextureID);

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int iCmdShow)
{
	WNDCLASS wc;
	MSG msg;

	file=lpCmdLine;

	wc.style=CS_VREDRAW|CS_HREDRAW|CS_OWNDC;
	wc.lpfnWndProc=WndProc;
	wc.cbClsExtra=0;
	wc.cbWndExtra=0;
	wc.hInstance=hInstance;
	wc.hIcon=LoadIcon(NULL, IDI_WINLOGO);
	wc.hCursor=LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground=GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName=NULL;
	wc.lpszClassName=szAppName;

	RegisterClass(&wc);

	WindowRect.left=0;
	WindowRect.right=Width;
	WindowRect.top=0;
	WindowRect.bottom=Height;

	AdjustWindowRect(&WindowRect, WS_OVERLAPPEDWINDOW, FALSE);

	hWnd=CreateWindow(szAppName, szAppName, WS_OVERLAPPEDWINDOW|WS_CLIPSIBLINGS|WS_CLIPCHILDREN, CW_USEDEFAULT, CW_USEDEFAULT, WindowRect.right-WindowRect.left, WindowRect.bottom-WindowRect.top, NULL, NULL, hInstance, NULL);

	ShowWindow(hWnd, SW_SHOW);
	SetForegroundWindow(hWnd);

	if(!Create())
		Done=TRUE;

	if(Done==FALSE)
		Resize(Width, Height);

	if(Done==FALSE)
		if(!Init())
			Done=TRUE;

	while(!Done)
	{
		if(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			if(msg.message==WM_QUIT)
				Done=TRUE;
			else
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
		else
		{
			Render();
			SwapBuffers(hDC);
		}
	}

	Destroy();
	DestroyWindow(hWnd);
	return msg.wParam;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg)
	{
		case WM_CREATE:
			break;

		case WM_CLOSE:
			PostQuitMessage(0);
			break;

		case WM_DESTROY:
			break;

		case WM_SIZE:
			Resize(LOWORD(lParam), HIWORD(lParam));
			break;

		case WM_KEYDOWN:
			switch(wParam)
			{
				case VK_ESCAPE:
					PostQuitMessage(0);
					break;

				default:
					break;
			}
			break;
	}
	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

void Render(void)
{
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	glPushMatrix();

	glPopMatrix();
}

BOOL Init(void)
{
	char infile[255], outfile[255];

	if(!IsExtensionSupported("GL_ARB_texture_compression"))
	{
		MessageBox(hWnd, "GL_ARB_texture_compression is not supported.", "OpenGL", MB_OK);
		return 0;
	}
	else
		glGetCompressedTexImageARB=(PFNGLGETCOMPRESSEDTEXIMAGEARBPROC)wglGetProcAddress("glGetCompressedTexImageARB");

	if(!IsExtensionSupported("GL_EXT_texture_compression_s3tc"))
	{
		MessageBox(hWnd, "GL_EXT_texture_compression_s3tc is not supported.", "OpenGL", MB_OK);
		return 0;
	}

	sprintf(infile, "%s.tga", file);
	compressed_texture=TGA_UploadImage(infile);

	sprintf(outfile, "%s.gtc", file);
	GTC_WriteImage(outfile, compressed_texture);

	return FALSE;
}

BOOL Create(void)
{
	PIXELFORMATDESCRIPTOR pfd;
	unsigned int PixelFormat;

	pfd.nSize=sizeof(PIXELFORMATDESCRIPTOR);
	pfd.nVersion=1;
	pfd.dwFlags=PFD_DRAW_TO_WINDOW|PFD_SUPPORT_OPENGL|PFD_DOUBLEBUFFER;
	pfd.iPixelType=PFD_TYPE_RGBA;
	pfd.cColorBits=32;
	pfd.cDepthBits=24;

	hDC=GetDC(hWnd);

	if(!(PixelFormat=ChoosePixelFormat(hDC, &pfd)))
	{
		MessageBox(hWnd, "ChoosePixelFormat Failed.", "ERROR", MB_OK);
		return FALSE;
	}

	if(!SetPixelFormat(hDC, PixelFormat, &pfd))
	{
		MessageBox(hWnd, "SetPixelFormat Failed.", "ERROR", MB_OK);
		return FALSE;
	}

	if(!(hRC=wglCreateContext(hDC)))
	{
		MessageBox(hWnd, "wglCreateContext Failed.", "ERROR", MB_OK);
		return FALSE;
	}

	if(!wglMakeCurrent(hDC, hRC))
	{
		MessageBox(hWnd, "wglMakeCurrent Failed.", "ERROR", MB_OK);
		return FALSE;
	}

	return TRUE;
}

void Destroy(void)
{
	wglMakeCurrent(NULL, NULL);
	wglDeleteContext(hRC);
	ReleaseDC(hWnd, hDC);
}

void Resize(int Width, int Height)
{
	if(Height==0)
		Height=1;

	glViewport(0,0, Width, Height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(90.0f, Width/Height, 0.001f, 400.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

BOOL IsExtensionSupported(const char *extension)
{
	const unsigned char *extensions=NULL, *start, *where, *terminator;

	where=strchr(extension, ' ');

	if(where||*extension=='\0')
		return FALSE;

	extensions=glGetString(GL_EXTENSIONS);
	start=extensions;

	for(;;)
	{
		where=strstr(start, extension);

		if(!where)
			break;

		terminator=where+strlen(extension);

		if(where==start||*(where-1)==' ')
			if(*terminator==' '||*terminator=='\0')
				return TRUE;

		start=terminator;
	}

	return FALSE;
}

BOOL TGA_GetData(char *file, TGA_t *TGA)
{
	FILE *stream;
	int size;

	if(!(stream=fopen(file, "rb")))
		return FALSE;

	fread(&TGA->ID, sizeof(unsigned char), 1, stream);
	fread(&TGA->ColorMapType, sizeof(unsigned char), 1, stream);
	fread(&TGA->ImageType, sizeof(unsigned char), 1, stream);
	fseek(stream, 5, SEEK_CUR);
	fread(&TGA->ImageSpec, sizeof(char), 10, stream);

	if(TGA->ID!=0)
		return FALSE;

	if(TGA->ColorMapType!=0)
		return FALSE;

	if(TGA->ImageType!=2)
		if(TGA->ImageType!=3)
			return FALSE;

	size=TGA->ImageSpec.Width*TGA->ImageSpec.Height;

	switch(TGA->ImageSpec.Depth)
	{
		case 32:
			TGA->Data=malloc(size*4);
			fread(TGA->Data, sizeof(unsigned char), size*4, stream);
			break;

		case 24:
			TGA->Data=malloc(size*3);
			fread(TGA->Data, sizeof(unsigned char), size*3, stream);
			break;

		case 8:
			TGA->Data=malloc(size);
			fread(TGA->Data, sizeof(unsigned char), size, stream);
			break;

		default:
			break;
	}

	fclose(stream);

	return TRUE;
}

unsigned int TGA_UploadImage(char *file)
{
	unsigned int TextureID;
	TGA_t TGA;

	if(!TGA_GetData(file, &TGA))
		return 0;

	switch(TGA.ImageSpec.Depth)
	{
		case 32:
			glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
			glGenTextures(1, &TextureID);
			glBindTexture(GL_TEXTURE_2D, TextureID);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_NEAREST);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			gluBuild2DMipmaps(GL_TEXTURE_2D, GL_COMPRESSED_RGBA_S3TC_DXT5_EXT, TGA.ImageSpec.Width, TGA.ImageSpec.Height, GL_BGRA_EXT, GL_UNSIGNED_BYTE, TGA.Data);
			break;

		case 24:
			glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
			glGenTextures(1, &TextureID);
			glBindTexture(GL_TEXTURE_2D, TextureID);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_NEAREST);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			gluBuild2DMipmaps(GL_TEXTURE_2D, GL_COMPRESSED_RGB_ARB, TGA.ImageSpec.Width, TGA.ImageSpec.Height, GL_BGR_EXT, GL_UNSIGNED_BYTE, TGA.Data);
			break;

		case 8:
			glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
			glGenTextures(1, &TextureID);
			glBindTexture(GL_TEXTURE_2D, TextureID);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_NEAREST);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			gluBuild2DMipmaps(GL_TEXTURE_2D, GL_COMPRESSED_ALPHA_ARB, TGA.ImageSpec.Width, TGA.ImageSpec.Height, GL_ALPHA, GL_UNSIGNED_BYTE, TGA.Data);
			break;

		default:
			break;
	}

	free(TGA.Data);

	width=TGA.ImageSpec.Width;
	height=TGA.ImageSpec.Height;
	return TextureID;
}

void GTC_WriteImage(char *file, unsigned int TextureID)
{
	FILE *stream;
	GTC_t GTC;

	glBindTexture(GL_TEXTURE_2D, TextureID);
	glEnable(GL_TEXTURE_2D);

	GTC.ID=GTC_ID;
	glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &GTC.Width);
	glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT, &GTC.Height);
	glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_INTERNAL_FORMAT, &GTC.Format);
	glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_COMPRESSED_IMAGE_SIZE_ARB, &GTC.Size);

	GTC.Data=malloc(GTC.Size);
	ZeroMemory(GTC.Data, GTC.Size);
	glGetCompressedTexImageARB(GL_TEXTURE_2D, 0, GTC.Data);

	if((stream=fopen(file, "wb"))==NULL)
		return;

	fwrite(&GTC.ID, sizeof(int), 1, stream);
	fwrite(&GTC.Width, sizeof(unsigned int), 1, stream);
	fwrite(&GTC.Height, sizeof(unsigned int), 1, stream);
	fwrite(&GTC.Format, sizeof(unsigned int), 1, stream);
	fwrite(&GTC.Size, sizeof(unsigned int), 1, stream);
	fwrite(GTC.Data, sizeof(unsigned char), GTC.Size, stream);

	fclose(stream);
	free(GTC.Data);
}